import logging
from typing import Optional, Union

from meross_iot.controller.mixins.toggle import ToggleMixin, ToggleXMixin
from meross_iot.model.enums import Namespace, LightMode
from meross_iot.model.plugin.light import LightInfo
from meross_iot.model.typing import RgbTuple
from meross_iot.utilities.conversion import rgb_to_int

_LOGGER = logging.getLogger(__name__)


class LightMixin(object):
    """
    Mixin class that enables light control.
    """
    _execute_command: callable
    check_full_update_done: callable

    # async_handle_update: Callable[[Namespace, dict], Awaitable]

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)

        # Dictionary keeping the status for every channel
        self._channel_light_status = {}

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False

        if namespace == Namespace.CONTROL_LIGHT:
            _LOGGER.debug(f"{self.__class__.__name__} handling push notification for namespace {namespace}")
            payload = data.get('light')
            if payload is None:
                _LOGGER.error(f"{self.__class__.__name__} could not find 'light' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                # Update the status of every channel that has been reported in this push
                # notification.
                c = payload['channel']
                self._update_channel_status(channel=c,
                                            rgb=payload.get('rgb'),
                                            luminance=payload.get('luminance'),
                                            temperature=payload.get('temperature'))
                locally_handled = True

        # Always call the parent handler when done with local specific logic. This gives the opportunity to all
        # ancestors to catch all events.
        parent_handled = await super().async_handle_push_notification(namespace=namespace, data=data)
        return locally_handled or parent_handled

    async def async_handle_update(self, namespace: Namespace, data: dict) -> bool:
        _LOGGER.debug(f"Handling {self.__class__.__name__} mixin data update.")
        locally_handled = False
        if namespace == Namespace.SYSTEM_ALL:
            light_data = data.get('all', {}).get('digest', {}).get('light', [])
            self._update_channel_status(channel=light_data.get('channel'),
                                        rgb=light_data.get('rgb'),
                                        luminance=light_data.get('luminance'),
                                        temperature=light_data.get('temperature'),
                                        onoff=light_data.get('onoff'))
            locally_handled = True
        super_handled = await super().async_handle_update(namespace=namespace, data=data)
        return super_handled or locally_handled

    def _supports_mode(self, mode: LightMode, channel: int = 0) -> bool:
        capacity = self.abilities.get(Namespace.CONTROL_LIGHT.value, {}).get('capacity')
        if capacity is None:
            return False

        return (capacity & mode.value) == mode.value

    def _update_channel_status(self,
                               channel: int = 0,
                               rgb: Union[int, RgbTuple] = None,
                               luminance: int = -1,
                               temperature: int = -1,
                               onoff: int = None) -> None:
        channel_info = self._channel_light_status.get(channel)
        if channel_info is None:
            channel_info = LightInfo()
            self._channel_light_status[channel] = channel_info

        channel_info.update(rgb=rgb, luminance=luminance, temperature=temperature, onoff=onoff)

    async def async_set_light_color(self,
                                    channel: int = 0,
                                    onoff: Optional[bool] = None,
                                    rgb: Optional[RgbTuple] = None,
                                    luminance: Optional[int] = None,
                                    temperature: Optional[int] = None,
                                    timeout: Optional[float] = None,
                                    *args,
                                    **kwargs) -> None:
        """
        Controls the light color of the given bulb. Please note that the __onoff parameter is ignored if the
        device supports Toggle or ToggleX operations__.

        :param channel: channel to control (for bulbs it's usually 0)
        :param onoff: when True, the device will be turned on, when false, it will turned off. This parameter is ignored
                      if the operating device must be controlled via ToggleX or Toggle command.
        :param rgb: (red,green,blue) tuple, where each color is an integer from 0-to-255
        :param luminance: Light intensity (at least on MSL120). Varies from 0 to 100
        :param temperature: Light temperature. Can be used when rgb is not specified.

        :return: None
        """
        if rgb is not None and temperature is not None:
            _LOGGER.error("You are trying to set both RGB and luminance values for this bulb. It won't work!")

        # Prepare a basic command payload
        light_payload = {}

        # For some reason, not all the Meross Devices do offer 'onoff' control attribute. For instance, the
        # light of the smart humidifier does require this parameter to be set, while the MSL120 requires
        # Toggle/ToggleX usage. For this reason, we'll only set the onoff value when Toggle/ToggleX is unavailable.
        is_on = self.get_light_is_on(channel=channel)

        # Handle "turn off" case for devices which implement togglex and toggle abilities.
        # In such case, we need to turn off the device explicitly. In all other cases, we can just send
        # the CONTROL_LIGHT command (which turns on the light).
        if (isinstance(self, ToggleMixin) or isinstance(self, ToggleXMixin)) and onoff == False and is_on != onoff:
            await self.async_turn_off(channel=channel, timeout=timeout)
        # In case the device does not support toggle/togglex, and the user explicitly asket to turn it on/off,
        # we must set the appropriate value for the light_payload.
        elif not isinstance(self, ToggleMixin) and not isinstance(self, ToggleXMixin) and onoff is not None and onoff != is_on:
            if onoff is not None:
                light_payload['onoff'] = 1 if onoff else 0
            elif self._channel_light_status[channel].is_on is not None:
                light_payload['onoff'] = 1 if self._channel_light_status[channel].is_on else 0

        if self._supports_mode(LightMode.MODE_RGB) and rgb is not None:
            # Convert the RGB to integer
            color = rgb_to_int(rgb)
            light_payload['rgb'] = color
            light_payload['capacity'] = light_payload.get('capacity',0) | LightMode.MODE_RGB.value

        if self._supports_mode(LightMode.MODE_LUMINANCE) and luminance is not None:
            light_payload['luminance'] = luminance
            light_payload['capacity'] = light_payload.get('capacity',0) | LightMode.MODE_LUMINANCE.value

        if self._supports_mode(LightMode.MODE_TEMPERATURE) and temperature is not None:
            light_payload['temperature'] = temperature
            light_payload['capacity'] = light_payload.get('capacity',0) | LightMode.MODE_TEMPERATURE.value

        # Only proceed sending the command if any attribute of the light payload was previously computed
        if light_payload != {}:
            light_payload.update({
                    'channel': channel,
                    'gradual': 0
                })
            payload = {
                'light': light_payload
            }

            #payload['light']['capacity'] = mode
            await self._execute_command(method='SET',
                                        namespace=Namespace.CONTROL_LIGHT,
                                        payload=payload,
                                        timeout=timeout)

        # If the command was ok, immediately update the local state.
        self._update_channel_status(channel, rgb=rgb, luminance=luminance, temperature=temperature, onoff=onoff)

    def get_supports_rgb(self, channel: int = 0) -> bool:
        """
        Tells if the current device supports RGB capability

        :param channel: channel to get info from, defaults to 0

        :return: True if the current device supports RGB color, False otherwise.
        """
        return self._supports_mode(LightMode.MODE_RGB, channel=channel)

    def get_supports_luminance(self, channel: int = 0) -> bool:
        """
        Tells if the current device supports luminance capability

        :param channel: channel to get info from, defaults to 0

        :return: True if the current device supports luminance mode, False otherwise.
        """
        return self._supports_mode(LightMode.MODE_LUMINANCE, channel=channel)

    def get_supports_temperature(self, channel: int = 0) -> bool:
        """
        Tells if the current device supports temperature color capability

        :param channel: channel to get info from, defaults to 0

        :return: True if the current device supports temperature mode, False otherwise.
        """
        return self._supports_mode(LightMode.MODE_TEMPERATURE, channel=channel)

    def get_rgb_color(self, channel=0, *args, **kwargs) -> Optional[RgbTuple]:
        """
        Returns the current RGB configuration of the device.

        :param channel: channel to control, defaults to 0 (bulbs generally have only one channel)

        :return: a Tuple containing three integer 8bits values (red, green, blue)
        """
        self.check_full_update_done()
        info = self._channel_light_status.get(channel)
        if info is None:
            return None
        return info.rgb_tuple

    def get_luminance(self, channel=0, *args, **kwargs) -> Optional[int]:
        """
        Returns the current brightness intensity of the bulb

        :param channel: channel to control, defaults to 0 (bulbs generally have only one channel)

        :return: an integer value from 0 to 100
        """
        self.check_full_update_done()
        info = self._channel_light_status.get(channel)
        if info is None:
            return None
        return info.luminance

    def get_color_temperature(self, channel=0, *args, **kwargs) -> Optional[int]:
        """
        Returns the current color_temperature

        :param channel: channel to control, defaults to 0 (bulbs generally have only one channel)

        :return: an integer value from 0 to 100
        """
        self.check_full_update_done()
        info = self._channel_light_status.get(channel)
        if info is None:
            return None
        return info.temperature

    def get_light_is_on(self, channel=0, *args, **kwargs) -> Optional[bool]:
        """
        Returns True if the light is ON, False otherwise.

        :param channel: channel to control, defaults to 0 (bulbs generally have only one channel)

        :return: current onoff state
        """
        # For some reason, Meross devices that support ToggleX and Toggle abilities, do not expose onoff
        # state within light status. In that case, we return the channel status.
        self.check_full_update_done()
        if isinstance(self, ToggleXMixin) or isinstance(self, ToggleMixin):
            return self.is_on(channel=channel)

        info = self._channel_light_status.get(channel)
        if info is None:
            return None
        return info.is_on

    async def async_turn_on(self, channel=0, *args, **kwargs) -> None:
        if isinstance(self, ToggleXMixin) or isinstance(self, ToggleMixin):
            await super().async_turn_on(channel=channel, *args, **kwargs)
        else:
            await self.async_set_light_color(channel=channel, onoff=True)

    async def async_turn_off(self, channel=0, *args, **kwargs) -> None:
        if isinstance(self, ToggleXMixin) or isinstance(self, ToggleMixin):
            await super().async_turn_off(channel=channel, *args, **kwargs)
        else:
            await self.async_set_light_color(channel=channel, onoff=False)



import logging
from typing import Optional

from meross_iot.model.enums import Namespace

_LOGGER = logging.getLogger(__name__)


class RollerShutterTimerMixin:
    _execute_command: callable
    check_full_update_done: callable
    uuid: str

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)
        self._roller_shutter_state_by_channel = {}
        self._roller_shutter_position_by_channel = {}

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False

        if namespace == Namespace.ROLLER_SHUTTER_STATE:
            _LOGGER.debug(f"{self.__class__.__name__} handling push notification for namespace "
                          f"{namespace}")
            payload = data.get('state')
            if payload is None:
                _LOGGER.error(f"{self.__class__.__name__} could not find 'state' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                # The roller shutter timer state push notification contains an object for every channel handled by the
                # device
                for roller_shutter in payload:
                    channel_index = roller_shutter['channel']
                    state = roller_shutter['state'] # open (position=100, state=1), close (position=0, state=2), stop (position=-1, state=0)
                    self._roller_shutter_state_by_channel[channel_index] = state
                    locally_handled = True
                    print(f"Push Notification - State: {state}")
                    
        elif namespace == Namespace.ROLLER_SHUTTER_POSITION:
            _LOGGER.debug(f"{self.__class__.__name__} handling push notification for namespace "
                          f"{namespace}")
            payload = data.get('position')
            if payload is None:
                _LOGGER.error(f"{self.__class__.__name__} could not find 'position' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                # The roller shutter timer position push notification contains an object for every channel handled by the
                # device
                for roller_shutter in payload:
                    channel_index = roller_shutter['channel']
                    position = roller_shutter['position'] # open (position=100, state=1), close (position=0, state=2), stop (position=-1, state=0)
                    self._roller_shutter_position_by_channel[channel_index] = position
                    locally_handled = True
                    print(f"Push Notification - Position: {position}")

        # Always call the parent handler when done with local specific logic. This gives the opportunity to all
        # ancestors to catch all events.
        parent_handled = await super().async_handle_push_notification(namespace=namespace, data=data)
        return locally_handled or parent_handled

    async def async_handle_update(self, namespace: Namespace, data: dict) -> bool:
        # TODO: The device not send the status or position values, only sends as digest:
        # 'digest': {'triggerx': [], 'timerx': []}}
        _LOGGER.debug(f"Handling {self.__class__.__name__} mixin data update.")
        locally_handled = False
        '''if namespace == Namespace.SYSTEM_ALL:
            roller_shutter_data = data.get('all', {}).get('digest', {})#.get('rollerShutter', [])
            for roller_shutter in roller_shutter_data:
                channel_index = roller_shutter['channel']
                state = roller_shutter['state']
                self._roller_shutter_state_by_channel[channel_index] = state
            locally_handled = True'''   

        super_handled = await super().async_handle_update(namespace=namespace, data=data)
        return super_handled or locally_handled

    async def async_open(self, channel: int = 0, *args, **kwargs) -> None:
        """
        Operates the roller shutter: sends the open command.

        :param channel: channel to operate: defaults to 0

        :return: None
        """
        await self._async_operate(position=100, channel=channel, *args, **kwargs)

    async def async_stop(self, channel: int = 0, *args, **kwargs) -> None:
        """
        Operates the roller shutter: sends the stop command.

        :param channel: channel to operate: defaults to 0

        :return: None
        """
        await self._async_operate(position=-1, channel=channel, *args, **kwargs)

    async def async_close(self, channel: int = 0, *args, **kwargs) -> None:
        """
        Operates the roller shutter: sends the close command.

        :param channel: channel to operate: defaults to 0

        :return: None
        """
        await self._async_operate(position=0, channel=channel, *args, **kwargs)

    async def _async_operate(self, position: int, channel: int = 0, timeout: Optional[float] = None, *args, **kwargs) -> None:
        payload = {'position': {"position": position, "channel": channel}}
        await self._execute_command(method="SET",
                                    namespace=Namespace.ROLLER_SHUTTER_POSITION,
                                    payload=payload,
                                    timeout=timeout)
        # Respect to other devices, we don't assume the command was immediately successful as the
        # state/position change might take some time to take place.
        # self._roller_shutter_state_by_channel[channel] = state
        # self._roller_shutter_position_by_channel[channel] = position

    def get_status(self, channel: int = 0, *args, **kwargs) -> Optional[int]:
        """
        The current roller shutter status. Returns 1 if the given roller shutter is open, 2 if it is close, 0 if it is stop.

        :param channel: channel of which status is needed

        :return: 1 if the given roller shutter is opened, 2 if it is closed, 0 if it is stopped.
        """
        self.check_full_update_done()
        return self._roller_shutter_state_by_channel.get(channel)

    def get_position(self, channel: int = 0, *args, **kwargs) -> Optional[int]:
        """
        The current roller shutter position. Returns 100 if the given roller shutter is open, 0 if it is close, -1 if it is stop.

        :param channel: channel of which status is needed

        :return: 100 if the given roller shutter is opened, 0 if it is closed, -1 if it is stopped.
        """
        self.check_full_update_done()
        return self._roller_shutter_position_by_channel.get(channel)


import logging
from typing import Optional

from meross_iot.model.enums import Namespace

_LOGGER = logging.getLogger(__name__)


class RollerShutterMixin:
    _execute_command: callable
    check_full_update_done: callable
    uuid: str

    def __init__(self, device_uuid: str,
                 manager,
                 **kwargs):
        super().__init__(device_uuid=device_uuid, manager=manager, **kwargs)
        self.__state_by_channel = {}

    async def async_handle_push_notification(self, namespace: Namespace, data: dict) -> bool:
        locally_handled = False

        if namespace in (Namespace.ROLLER_SHUTTER_STATE, Namespace.ROLLER_SHUTTER_POSITION):
            _LOGGER.debug(f"{self.__class__.__name__} handling push notification for namespace "
                          f"{namespace}")
            payload = data.get('state')
            if payload is None:
                _LOGGER.error(f"{self.__class__.__name__} could not find 'state' attribute in push notification data: "
                              f"{data}")
                locally_handled = False
            else:
                # The door opener state push notification contains an object for every channel handled by the
                # device
                for shutter in payload:
                    channel_index = shutter['channel']
                    state = shutter.get('state')
                    self.__state_by_channel.get(channel_index, {}).update(state)
                    locally_handled = True

        # Always call the parent handler when done with local specific logic. This gives the opportunity to all
        # ancestors to catch all events.
        parent_handled = await super().async_handle_push_notification(namespace=namespace, data=data)
        return locally_handled or parent_handled

    async def async_handle_update(self, namespace: Namespace, data: dict) -> bool:
        _LOGGER.debug(f"Handling {self.__class__.__name__} mixin data update.")
        locally_handled = False
        """
        if namespace == Namespace.SYSTEM_ALL:
            doors_data = data.get('all', {}).get('digest', {}).get('garageDoor', [])
            for door in doors_data:
                channel_index = door['channel']
                state = door['open'] == 1
                self._door_open_state_by_channel[channel_index] = state
            locally_handled = True
        """
        super_handled = await super().async_handle_update(namespace=namespace, data=data)
        return super_handled or locally_handled

